import numpy as np
from tensorflow import keras
from tensorflow.keras import backend as K
from tensorflow.keras.datasets import fashion_mnist
from keras_drop_block import DropBlock2D


(x_train, y_train), (x_test, y_test) = fashion_mnist.load_data()

x_train = np.expand_dims(x_train.astype(K.floatx()) / 255, axis=-1)
x_test = np.expand_dims(x_test.astype(K.floatx()) / 255, axis=-1)

y_train, y_test = np.expand_dims(y_train, axis=-1), np.expand_dims(y_test, axis=-1)

train_num = round(x_train.shape[0] * 0.9)
x_train, x_valid = x_train[:train_num, ...], x_train[train_num:, ...]
y_train, y_valid = y_train[:train_num, ...], y_train[train_num:, ...]


def get_dropout_model():
    model = keras.models.Sequential()
    model.add(keras.layers.Dropout(input_shape=(28, 28, 1), rate=0.3, name='Input-Dropout'))
    model.add(keras.layers.Conv2D(filters=64, kernel_size=3, activation='relu', padding='same', name='Conv-1'))
    model.add(keras.layers.MaxPool2D(pool_size=2, name='Pool-1'))
    model.add(keras.layers.Dropout(rate=0.2, name='Dropout-1'))
    model.add(keras.layers.Conv2D(filters=32, kernel_size=3, activation='relu', padding='same', name='Conv-2'))
    model.add(keras.layers.MaxPool2D(pool_size=2, name='Pool-2'))
    model.add(keras.layers.Dropout(rate=0.2, name='Dropout-2'))
    model.add(keras.layers.Flatten(name='Flatten'))
    model.add(keras.layers.Dense(units=256, activation='relu', name='Dense'))
    model.add(keras.layers.Dropout(rate=0.2, name='Dense-Dropout'))
    model.add(keras.layers.Dense(units=10, activation='softmax', name='Softmax'))
    model.compile(
        optimizer='adam',
        loss='sparse_categorical_crossentropy',
        metrics=['accuracy'],
    )
    return model


dropout_model = get_dropout_model()
dropout_model.summary()
dropout_model.fit(
    x=x_train,
    y=y_train,
    epochs=10,
    validation_data=(x_valid, y_valid),
    callbacks=[keras.callbacks.EarlyStopping(monitor='val_acc', patience=2)]
)
dropout_score = dropout_model.evaluate(x_test, y_test)
print('Score of dropout:\t%.4f' % dropout_score[1])


def get_drop_block_model():
    model = keras.models.Sequential()
    model.add(DropBlock2D(input_shape=(28, 28, 1), block_size=7, keep_prob=0.8, name='Input-Dropout'))
    model.add(keras.layers.Conv2D(filters=64, kernel_size=3, activation='relu', padding='same', name='Conv-1'))
    model.add(keras.layers.MaxPool2D(pool_size=2, name='Pool-1'))
    model.add(DropBlock2D(block_size=5, keep_prob=0.8, name='Dropout-1'))
    model.add(keras.layers.Conv2D(filters=32, kernel_size=3, activation='relu', padding='same', name='Conv-2'))
    model.add(keras.layers.MaxPool2D(pool_size=2, name='Pool-2'))
    model.add(DropBlock2D(block_size=3, keep_prob=0.8, name='Dropout-2'))
    model.add(keras.layers.Flatten(name='Flatten'))
    model.add(keras.layers.Dense(units=256, activation='relu', name='Dense'))
    model.add(keras.layers.Dropout(rate=0.2, name='Dense-Dropout'))
    model.add(keras.layers.Dense(units=10, activation='softmax', name='Softmax'))
    model.compile(
        optimizer='adam',
        loss='sparse_categorical_crossentropy',
        metrics=['accuracy'],
    )
    return model


drop_block_model = get_drop_block_model()
drop_block_model.summary()
drop_block_model.fit(
    x=x_train,
    y=y_train,
    epochs=10,
    validation_data=(x_valid, y_valid),
    callbacks=[keras.callbacks.EarlyStopping(monitor='val_acc', patience=2)]
)
drop_block_score = drop_block_model.evaluate(x_test, y_test)
print('Score of dropout:\t%.4f' % dropout_score[1])
print('Score of DropBlock:\t%.4f' % drop_block_score[1])



import os
import random
import tempfile
import unittest

import numpy as np
from tensorflow import keras

from keras_drop_block import DropBlock1D


class TestDropBlock1D(unittest.TestCase):

    def test_training(self):
        input_layer = keras.layers.Input(shape=(100, 3))
        drop_block_layer = DropBlock1D(block_size=3, keep_prob=0.7)(input_layer)
        model = keras.models.Model(inputs=input_layer, outputs=drop_block_layer)
        model.compile(optimizer='adam', loss='mse', metrics={})
        model_path = os.path.join(tempfile.gettempdir(), 'keras_drop_block_%f.h5' % random.random())
        model.save(model_path)
        model = keras.models.load_model(
            model_path,
            custom_objects={'DropBlock1D': DropBlock1D},
        )
        model.summary()
        inputs = np.ones((1, 100, 3))
        outputs = model.predict(inputs)
        self.assertTrue(np.allclose(inputs, outputs))

        input_layer = keras.layers.Input(shape=(3, 100))
        drop_block_layer = DropBlock1D(block_size=3, keep_prob=0.7, data_format='channels_first')(input_layer)
        model = keras.models.Model(inputs=input_layer, outputs=drop_block_layer)
        model.compile(optimizer='adam', loss='mse', metrics={})
        model_path = os.path.join(tempfile.gettempdir(), 'keras_drop_block_%f.h5' % random.random())
        model.save(model_path)
        model = keras.models.load_model(
            model_path,
            custom_objects={'DropBlock1D': DropBlock1D},
        )
        model.summary()
        inputs = np.ones((1, 3, 100))
        outputs = model.predict(inputs)
        self.assertTrue(np.allclose(inputs, outputs))

    def test_mask_shape(self):
        input_layer = keras.layers.Input(shape=(100, 3))
        drop_block_layer = DropBlock1D(block_size=3, keep_prob=0.7)(input_layer, training=True)
        model = keras.models.Model(inputs=input_layer, outputs=drop_block_layer)
        model.compile(optimizer='adam', loss='mse', metrics={})
        model_path = os.path.join(tempfile.gettempdir(), 'keras_drop_block_%f.h5' % random.random())
        model.save(model_path)
        model = keras.models.load_model(
            model_path,
            custom_objects={'DropBlock1D': DropBlock1D},
        )
        model.summary()
        inputs = np.ones((1, 100, 3))
        outputs = model.predict(inputs)
        for i in range(3):
            print((outputs[0, :, i] > 0.0).astype(dtype='int32'))
        inputs = np.ones((1000, 100, 3))
        outputs = model.predict(inputs)
        keep_prob = 1.0 * np.sum(outputs > 0.0) / np.prod(np.shape(outputs))
        print(keep_prob)
        self.assertTrue(0.65 < keep_prob < 0.8, keep_prob)

        input_layer = keras.layers.Input(shape=(3, 100))
        drop_block_layer = DropBlock1D(block_size=3, keep_prob=0.7,
                                       data_format='channels_first')(input_layer, training=True)
        model = keras.models.Model(inputs=input_layer, outputs=drop_block_layer)
        model.compile(optimizer='adam', loss='mse', metrics={})
        model_path = os.path.join(tempfile.gettempdir(), 'keras_drop_block_%f.h5' % random.random())
        model.save(model_path)
        model = keras.models.load_model(
            model_path,
            custom_objects={'DropBlock1D': DropBlock1D},
        )
        model.summary()
        inputs = np.ones((1, 3, 100))
        outputs = model.predict(inputs)
        for i in range(3):
            print((outputs[0, i, :] > 0.0).astype(dtype='int32'))
        inputs = np.ones((1000, 3, 100))
        outputs = model.predict(inputs)
        keep_prob = 1.0 * np.sum(outputs > 0.0) / np.prod(np.shape(outputs))
        print(keep_prob)
        self.assertTrue(0.65 < keep_prob < 0.8, keep_prob)

    def test_sync_channels(self):
        input_layer = keras.layers.Input(shape=(100, 3))
        drop_block_layer = DropBlock1D(block_size=3, keep_prob=0.7, sync_channels=True)(input_layer, training=True)
        model = keras.models.Model(inputs=input_layer, outputs=drop_block_layer)
        model.compile(optimizer='adam', loss='mse', metrics={})
        model_path = os.path.join(tempfile.gettempdir(), 'keras_drop_block_%f.h5' % random.random())
        model.save(model_path)
        model = keras.models.load_model(
            model_path,
            custom_objects={'DropBlock1D': DropBlock1D},
        )
        model.summary()
        inputs = np.ones((1, 100, 3))
        outputs = model.predict(inputs)
        for i in range(1, 3):
            self.assertTrue(np.allclose(outputs[0, :, 0], outputs[0, :, i]))
        inputs = np.ones((1000, 100, 3))
        outputs = model.predict(inputs)
        keep_prob = 1.0 * np.sum(outputs > 0.0) / np.prod(np.shape(outputs))
        print(keep_prob)
        self.assertTrue(0.65 < keep_prob < 0.8, keep_prob)



from tensorflow import keras
from tensorflow.keras import backend as K


class DropBlock1D(keras.layers.Layer):
    """See: https://arxiv.org/pdf/1810.12890.pdf"""

    def __init__(self,
                 block_size,
                 keep_prob,
                 sync_channels=False,
                 data_format=None,
                 **kwargs):
        """Initialize the layer.

        :param block_size: Size for each mask block.
        :param keep_prob: Probability of keeping the original feature.
        :param sync_channels: Whether to use the same dropout for all channels.
        :param data_format: 'channels_first' or 'channels_last' (default).
        :param kwargs: Arguments for parent class.
        """
        super(DropBlock1D, self).__init__(**kwargs)
        self.block_size = block_size
        self.keep_prob = keep_prob
        self.sync_channels = sync_channels
        self.data_format = data_format
        self.supports_masking = True
        self.seq_len = self.ones = self.zeros = None

    def build(self, input_shape):
        if self.data_format == 'channels_first':
            self.seq_len = input_shape[-1]
        else:
            self.seq_len = input_shape[1]
        self.ones = K.ones(self.seq_len, name='ones')
        self.zeros = K.zeros(self.seq_len, name='zeros')
        super().build(input_shape)

    def get_config(self):
        config = {'block_size': self.block_size,
                  'keep_prob': self.keep_prob,
                  'sync_channels': self.sync_channels,
                  'data_format': self.data_format}
        base_config = super(DropBlock1D, self).get_config()
        return dict(list(base_config.items()) + list(config.items()))

    def compute_mask(self, inputs, mask=None):
        return mask

    def compute_output_shape(self, input_shape):
        return input_shape

    def _get_gamma(self):
        """Get the number of activation units to drop"""
        feature_dim = K.cast(self.seq_len, K.floatx())
        block_size = K.constant(self.block_size, dtype=K.floatx())
        return ((1.0 - self.keep_prob) / block_size) * (feature_dim / (feature_dim - block_size + 1.0))

    def _compute_valid_seed_region(self):
        positions = K.arange(self.seq_len)
        half_block_size = self.block_size // 2
        valid_seed_region = K.switch(
            K.all(
                K.stack(
                    [
                        positions >= half_block_size,
                        positions < self.seq_len - half_block_size,
                    ],
                    axis=-1,
                ),
                axis=-1,
            ),
            self.ones,
            self.zeros,
        )
        return K.expand_dims(K.expand_dims(valid_seed_region, axis=0), axis=-1)

    def _compute_drop_mask(self, shape):
        mask = K.random_binomial(shape, p=self._get_gamma())
        mask *= self._compute_valid_seed_region()
        mask = keras.layers.MaxPool1D(
            pool_size=self.block_size,
            padding='same',
            strides=1,
            data_format='channels_last',
        )(mask)
        return 1.0 - mask

    def call(self, inputs, training=None):

        def dropped_inputs():
            outputs = inputs
            if self.data_format == 'channels_first':
                outputs = K.permute_dimensions(outputs, [0, 2, 1])
            shape = K.shape(outputs)
            if self.sync_channels:
                mask = self._compute_drop_mask([shape[0], shape[1], 1])
            else:
                mask = self._compute_drop_mask(shape)
            outputs = outputs * mask *\
                (K.cast(K.prod(shape), dtype=K.floatx()) / K.sum(mask))
            if self.data_format == 'channels_first':
                outputs = K.permute_dimensions(outputs, [0, 2, 1])
            return outputs

        return K.in_train_phase(dropped_inputs, inputs, training=training)


class DropBlock2D(keras.layers.Layer):
    """See: https://arxiv.org/pdf/1810.12890.pdf"""

    def __init__(self,
                 block_size,
                 keep_prob,
                 sync_channels=False,
                 data_format=None,
                 **kwargs):
        """Initialize the layer.

        :param block_size: Size for each mask block.
        :param keep_prob: Probability of keeping the original feature.
        :param sync_channels: Whether to use the same dropout for all channels.
        :param data_format: 'channels_first' or 'channels_last' (default).
        :param kwargs: Arguments for parent class.
        """
        super(DropBlock2D, self).__init__(**kwargs)
        self.block_size = block_size
        self.keep_prob = keep_prob
        self.sync_channels = sync_channels
        self.data_format = data_format
        self.supports_masking = True
        self.height = self.width = self.ones = self.zeros = None

    def build(self, input_shape):
        if self.data_format == 'channels_first':
            self.height, self.width = input_shape[2], input_shape[3]
        else:
            self.height, self.width = input_shape[1], input_shape[2]
        self.ones = K.ones((self.height, self.width), name='ones')
        self.zeros = K.zeros((self.height, self.width), name='zeros')
        super().build(input_shape)

    def get_config(self):
        config = {'block_size': self.block_size,
                  'keep_prob': self.keep_prob,
                  'sync_channels': self.sync_channels,
                  'data_format': self.data_format}
        base_config = super(DropBlock2D, self).get_config()
        return dict(list(base_config.items()) + list(config.items()))

    def compute_mask(self, inputs, mask=None):
        return mask

    def compute_output_shape(self, input_shape):
        return input_shape

    def _get_gamma(self):
        """Get the number of activation units to drop"""
        height, width = K.cast(self.height, K.floatx()), K.cast(self.width, K.floatx())
        block_size = K.constant(self.block_size, dtype=K.floatx())
        return ((1.0 - self.keep_prob) / (block_size ** 2)) *\
               (height * width / ((height - block_size + 1.0) * (width - block_size + 1.0)))

    def _compute_valid_seed_region(self):
        positions = K.concatenate([
            K.expand_dims(K.tile(K.expand_dims(K.arange(self.height), axis=1), [1, self.width]), axis=-1),
            K.expand_dims(K.tile(K.expand_dims(K.arange(self.width), axis=0), [self.height, 1]), axis=-1),
        ], axis=-1)
        half_block_size = self.block_size // 2
        valid_seed_region = K.switch(
            K.all(
                K.stack(
                    [
                        positions[:, :, 0] >= half_block_size,
                        positions[:, :, 1] >= half_block_size,
                        positions[:, :, 0] < self.height - half_block_size,
                        positions[:, :, 1] < self.width - half_block_size,
                    ],
                    axis=-1,
                ),
                axis=-1,
            ),
            self.ones,
            self.zeros,
        )
        return K.expand_dims(K.expand_dims(valid_seed_region, axis=0), axis=-1)

    def _compute_drop_mask(self, shape):
        mask = K.random_binomial(shape, p=self._get_gamma())
        mask *= self._compute_valid_seed_region()
        mask = keras.layers.MaxPool2D(
            pool_size=(self.block_size, self.block_size),
            padding='same',
            strides=1,
            data_format='channels_last',
        )(mask)
        return 1.0 - mask

    def call(self, inputs, training=None):

        def dropped_inputs():
            outputs = inputs
            if self.data_format == 'channels_first':
                outputs = K.permute_dimensions(outputs, [0, 2, 3, 1])
            shape = K.shape(outputs)
            if self.sync_channels:
                mask = self._compute_drop_mask([shape[0], shape[1], shape[2], 1])
            else:
                mask = self._compute_drop_mask(shape)
            outputs = outputs * mask *\
                (K.cast(K.prod(shape), dtype=K.floatx()) / K.sum(mask))
            if self.data_format == 'channels_first':
                outputs = K.permute_dimensions(outputs, [0, 3, 1, 2])
            return outputs

        return K.in_train_phase(dropped_inputs, inputs, training=training)
